/*
 * mmtcpfwd v0.3-beta - Port forwarder superserver with DoS protection
 * Copyright (C) 2000  Matthew Mondor
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Original author: Matthew Mondor, mmondor@linuxguru.net
 */

/*
 * Uses secure input functions which should not overflow. Also uses security
 * features as logging, and limits global number of connections, as well
 * as how many connections are allowed per IP address.
 * Uses extensive error checking so that it should never segfault.
 * 
 * Upon too many refused connections exceeded by an ip, it is logged and an
 * ipchains rule is added to DENY any packets comming from that ip address.
 * To be able to do this, it internally starts a thread running as root, which
 * only can accept request from an internal pipe, so that IP addresses may
 * be denied. Other threads handling tcp/ip and serving clients run as wanted
 * userid and groupid.
 * 
 * It thus should be very secure. To make it even less vulnerable, the same
 * safe I/O functions are used for communication between the daemons running as
 * nobody and the root service process. The pipe write filedescriptor is never
 * passed to the threads servicing the clients, that also run as nobody.
 * 
 * Moreover, if the maximum allowed connections per IP is set to 0, it thus
 * immediately sets a DENY rule for any IP attempting a connection. An
 * independant process is started for each port it listens to, and pthreads
 * are used to service multiple connections at once on each. This is good for
 * ICMP
 *
 * TODO:
 * - Find a way to close all current connections to an IP which is to
 *   be DENYed.
 * - Add encryption support for this to serve as a  proxy, with
 *   authentication from a client, with uid check would be required
 */




// INCLUDES
// ========

#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>

#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <syslog.h>

#include <pthread.h>
#include <signal.h>
#include <sys/poll.h>
#include <time.h>

#include "mmfd.h"
#include "mmlist.h"




// DEFINITIONS
// ===========

// Useful when using pipe() generated fd pairs
#define P_READ   0
#define P_WRITE  1




// STRUCTURES
// ==========

/* This structure is a communication node between the parent
 * and the multiple threads.
 */
typedef struct clistnode
{
  lnode node;

  // Important info on the client
  struct sockaddr_in client;
  LONG fromport;
  int socket;

  // To control the connection attempts on a per-ip basis
  LONG connections, refused;

  // Global information pointers
  list *clist;
  pthread_mutex_t *clist_lock;
  pthread_mutex_t *rootpipe_lock;
  BOOL resolve, transproxy;
  LONG uid, gid;
  BYTE **closed_status;

  // IP and port to forward connections to
  struct sockaddr_in toserver;
  LONG toport;

  // Inactivity timeout for connections
  int timeout;

  // Assigned connection ID
  ULONG ID;
}
clistnode;

// Information returned by start_rootserver()
typedef struct rootdata
{
  int pid;
  int filedes[2];
}
rootdata;

/* Information to be passwd to each fork()ed thread
 * (one per tcp port)
 */
typedef struct flistnode
{
  lnode node;
  rootdata *rdata;
  pthread_mutex_t *rootpipe_lock;
  LONG port, maxperip, maxips, maxrej, toport, mstimeout, uid, gid;
  BYTE tohost[80];
  BOOL resolve, transproxy;
  int pid;			// Set by start_fwdserver
}
flistnode;




// PROTOTYPES
// ==========

int main (int, char **);
int start_fwdserver (flistnode *);
BOOL make_daemon (VOID);
BOOL write_pid (BYTE *);
LONG readconfline (BYTE *, FILE *, LONG, LONG, BYTE **);
rootdata *start_rootserver (VOID);
VOID phandleclient (VOID *);
int handleclient (int, clistnode *);




// GLOBAL VARIABLES
// ================

static pthread_mutex_t rootpipe_lock;
static list *flist;
static BYTE version[] = "\r\n+++ mmtcpfwd 0.3-beta (Matthew Mondor)\r\n";




// MAIN PROGRAM
// ============

int
main (int argc, char *argv[])
{
  struct rootdata *rdata;
  flistnode *flnode;
  int ret;
  LONG uid, gid, cols;
  BYTE line[1024], *array[16], defcfg[] = "/etc/mmtcpfwd.conf", *cfg = defcfg;
  FILE *fh;


  puts (version);


  // Initialize our global rootpipe_lock
  pthread_mutex_init (&rootpipe_lock, NULL);

  openlog ("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);

  // Start our rootserver thread
  syslog (LOG_NOTICE, "Starting RootServer thread");
  if (!(rdata = start_rootserver ()))
    {
      syslog (LOG_NOTICE, "Could not start RootServer thread");
      puts (" * Cannot start rootserver thread");
      exit (-1);
    }

  // Read our config file, binding ports and starting a thread for each
  if (argc == 2)
    cfg = argv[1];
  if (argc < 3)
    {
      if ((fh = fopen (cfg, "r")))
	{

	  // Get uid and gid from config file
	  if ((readconfline (line, fh, 2, 2, array)) == 2)
	    {
	      uid = atol (array[0]);
	      gid = atol (array[1]);
	    }
	  else
	    {
	      puts
		(" * Wrong configuration file, first line should have uid gid columns\r\n");
	      close (fh);
	      exit (-1);
	    }

	  // Now in a loop read all port forwarding entries
	  if ((flist = createlist (sizeof (flistnode))))
	    {
	      ret = 0;
	      while ((cols = readconfline (line, fh, 4, 9, array)) != -1)
		{
		  if (cols)
		    {
		      if (cols == 4)
			{
			  if (!(atol (array[1])))
			    {
			      if ((flnode = (flistnode *) appendnode (flist)))
				{
				  // This is a fake service we deny IPs to
				  flnode->rdata = rdata;
				  flnode->rootpipe_lock = &rootpipe_lock;
				  flnode->port = atol (array[0]);
				  if (atol (array[1]))
				    flnode->resolve = TRUE;
				  else
				    flnode->resolve = FALSE;
				  flnode->maxperip = 0;
				  flnode->maxips = atol (array[3]);
				  flnode->maxrej = 0;
				  flnode->tohost[0] = 0;
				  flnode->toport = 0;
				  flnode->mstimeout = 0;
				  flnode->uid = uid;
				  flnode->gid = gid;
				  flnode->pid = 0;

				  if (!(start_fwdserver (flnode)))
				    {
				      printf
					(" * Error stating a fwd thread on port %s\r\n",
					 flnode->port);
				      ret = -1;
				      break;
				    }
				}
			      else
				{
				  puts (" * Error allocating memory");
				  ret = -1;
				  break;
				}
			    }
			  else
			    {
			      puts
				(" * Error parsing config file (fake server should have 0/IP");
			      ret = -1;
			      break;
			    }
			}
		      else if (cols == 9)
			{
			  if ((flnode = (flistnode *) appendnode (flist)))
			    {
			      // This is an actual service we forward
			      flnode->rdata = rdata;
			      flnode->rootpipe_lock = &rootpipe_lock;
			      flnode->port = atol (array[0]);
			      if (atol (array[1]))
				flnode->resolve = TRUE;
			      else
				flnode->resolve = FALSE;
			      flnode->maxperip = atol (array[2]);
			      flnode->maxips = atol (array[3]);
			      flnode->maxrej = atol (array[4]);
			      strncpy (flnode->tohost, array[5], 78);
			      flnode->toport = atol (array[6]);
			      if (atol (array[7]))
				flnode->transproxy = TRUE;
			      else
				flnode->transproxy = FALSE;
			      flnode->mstimeout = atol (array[8]);
			      flnode->uid = uid;
			      flnode->gid = gid;
			      flnode->pid = 0;

			      if (!(start_fwdserver (flnode)))
				{
				  printf
				    (" * Error starting a fwd thread on port %d\r\n",
				     flnode->port);
				  ret = -1;
				  break;
				}
			    }
			  else
			    {
			      puts (" * Error allocating memory");
			      ret = -1;
			      break;
			    }
			}
		      else
			{
			  puts (" * Error in configuration file entry");
			  ret = -1;
			  break;
			}
		    }
		}
	    }
	  if (ret == -1)
	    {
	      if (fh)
		fclose (fh);
	      if (flist)
		freelist (flist);
	      exit (ret);
	    }
	}
      else
	{
	  printf (" * Could not open config file (%s)\r\n\r\n", cfg);
	  exit (-1);
	}
    }
  else
    {
      puts ("USAGE: mmtcpfwd [<configfile>]\r\n");
      exit (-1);
    }

  exit (0);
}


// This is the actual main program which every thread execute for each port.
int
start_fwdserver (flistnode * fnode)
{
  struct sockaddr addr;
  struct sockaddr_in server, toserver, *saddr = (struct sockaddr_in *) &addr;
  struct hostent *he;
  struct in_addr *ip;

  pthread_mutex_t clist_lock;
  list *clist;

  socklen_t addrl;
  clistnode *clnode;

  int sock, msgsock, ret, pid, fd;
  pthread_t thread;
  pthread_attr_t threadattr;
  BYTE *tmp, ipaddr[20], line[1024];
  BOOL found, ok;
  ULONG ID = 0;

  // We keep this array and share it with all the pthreads handling connections
  BYTE *closed_status[] = {
    "STATUS_UNKNOWN",
    "CLIENT_CLOSED",
    "SERVICE_CLOSED",
    "CLIENT_READ_EOF",
    "SERVICE_READ_EOF",
    "CLIENT_WRITE_ERROR",
    "SERVICE_WRITE_ERROR",
    "SERVICE_CONNECT_ERROR",
    "CREATE_SOCKET_ERROR",
    "TRANSPROXY_BIND_ERROR"
  };


  pid = fork ();
  if (pid == 0)
    {

      // Child process, Detach...
      setsid ();
      chdir ("/");
      if ((fd = open ("/dev/null", O_RDWR)) != -1)
	{
	  dup2 (fd, 0);
	  dup2 (fd, 1);
	  dup2 (fd, 2);
	  if (fd > 2)
	    close (fd);
	}
      signal (SIGTTOU, SIG_IGN);
      signal (SIGTTIN, SIG_IGN);
      signal (SIGTSTP, SIG_IGN);
      umask (0);

      openlog ("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);

      // Here we setup our toserver sockaddr_in structure
      if (fnode->maxperip)
	{

	  syslog (LOG_NOTICE, "Starting forwarder on port %d", fnode->port);

	  // Resolve tohost to IP
	  if ((he = gethostbyname (fnode->tohost)))
	    {
	      ip = (struct in_addr *) he->h_addr_list[0];
	    }
	  else
	    {
	      syslog (LOG_NOTICE,
		      "Cannot resolve hostname to forward port %d",
		      fnode->port);
	      return (0);
	    }

	  // Setup toserver structure
	  memset (&toserver, 0, sizeof (struct sockaddr_in));
	  toserver.sin_family = AF_INET;
	  toserver.sin_addr.s_addr = ip->s_addr;
	  toserver.sin_port = htons ((UWORD) fnode->toport);

	}
      else
	{
	  syslog (LOG_NOTICE, "Starting fake service on port %d",
		  fnode->port);
	}


      // Prepare pthread attributes, as well as clientslist lock
      pthread_attr_init (&threadattr);
      pthread_attr_setdetachstate (&threadattr, PTHREAD_CREATE_DETACHED);
      pthread_mutex_init (&clist_lock, NULL);


      // Create clientslist
      if ((clist = createlist (sizeof (clistnode))))
	{

	  /* This section binds a port, listens for connections to it, and
	   * dispaches each client connecting to it to a multi-tasking client
	   * server thread
	   */
	  ret = 0;
	  if ((sock = socket (AF_INET, SOCK_STREAM, 0)) > -1)
	    {
	      memset (&server, 0, sizeof (server));
	      server.sin_family = AF_INET;
	      server.sin_addr.s_addr = INADDR_ANY;
	      server.sin_port = htons (fnode->port);
	      if (!
		  (bind
		   (sock, (struct sockaddr *) (void *) &server,
		    sizeof (server))))
		{

		  // Become non-root user and group (usually nobody/nogroup)
		  setgid (fnode->gid);
		  setuid (fnode->uid);

		  if (!(listen (sock, fnode->maxips)))
		    {

		      // Wait for client connections and handle them
		      while (TRUE)
			{
			  addrl = sizeof (struct sockaddr);
			  if ((msgsock = accept (sock, &addr, &addrl)) != -1)
			    {

			      /* If -c0 was specified, we simply immediately
			       * close connection and DENY this IP.
			       */
			      if (!fnode->maxperip)
				{
				  if ((tmp = inet_ntoa (saddr->sin_addr)))
				    {
				      // Send a warning message
				      fdputs (msgsock,
					      "Locator Succeeded\r\n");
				      // Get IP
				      strncpy (ipaddr, tmp, 19);
				      // Close socket
				      close (msgsock);
				      // DENY IP, send request locking rootpipe
				      pthread_mutex_lock
					(fnode->rootpipe_lock);
				      fdprintf (fnode->
						rdata->filedes[P_WRITE],
						"%s:%d\n", ipaddr,
						fnode->port);
				      pthread_mutex_unlock
					(fnode->rootpipe_lock);
				      // Continue main loop
				      continue;
				    }
				}

			      /* Obtain the IP address of the current client and check
			       * if existing in the list of connected clients. If so,
			       * ensures that the client does not exceed allowable
			       * connections to server, and increase the connections
			       * number for that client. If not, add a new node in the
			       * list for the new client, if we still have allowed
			       * connections left
			       */

			      // Lock the list so that no other process can access it
			      pthread_mutex_lock (&clist_lock);

			      /* Try to locate the IP of the new connection to the
			       * list of currently connected clients'
			       */
			      found = FALSE;
			      if ((clnode = (clistnode *) topnode (clist)))
				{
				  while (clnode)
				    {
				      if (clnode->client.sin_addr.s_addr ==
					  saddr->sin_addr.s_addr)
					{
					  found = TRUE;
					  break;
					}
				      clnode = (clistnode *) nextnode (clist);
				    }
				}
			      ok = FALSE;

			      if (found)
				{

				  /* The client is already connected. Make sure he is
				   * still allowed connections
				   */
				  if (clnode->connections < fnode->maxperip)
				    {

				      // This ip still allowed connections
				      clnode->connections++;
				      ok = TRUE;

				    }
				  else
				    {

				      /* The client attempts to exceed his allowed
				       * simultanious connections
				       */
				      clnode->refused++;
				      if (clnode->refused > fnode->maxrej)
					{

					  /* Client has exceeded the number of
					   * allowed failed connections attempts.
					   * He is DoSing. Log it, and DENY that
					   * address using ipchains. This is done
					   * by telling our rootserver to so it.
					   */
					  if (
					      (tmp =
					       inet_ntoa (clnode->
							  client.sin_addr)))
					    {
					      strncpy (ipaddr, tmp, 19);
					      close (msgsock);
					      msgsock = 0;
					      pthread_mutex_lock
						(fnode->rootpipe_lock);
					      fdprintf (fnode->rdata->filedes
							[P_WRITE], "%s:%d\n",
							ipaddr, fnode->port);
					      pthread_mutex_unlock
						(fnode->rootpipe_lock);
					    }

					}

				    }
				}
			      else
				{

				  /* If maximum total connections isn't reached, add
				   * a new node for new client
				   */
				  if ((totalnodes (clist)) < fnode->maxips)
				    {
				      if (
					  (clnode =
					   (clistnode *) appendnode (clist)))
					{
					  clnode->connections = 1;
					  ok = TRUE;
					}
				    }
				}

			      if (ok)
				{

				  /* We start thread to serve the client.
				   * also set parameters it requires.
				   */
				  clnode->client = *saddr;
				  clnode->socket = msgsock;
				  clnode->clist = clist;
				  clnode->clist_lock = &clist_lock;
				  clnode->closed_status = closed_status;
				  clnode->toserver = toserver;
				  clnode->fromport = fnode->port;
				  clnode->toport = fnode->toport;
				  clnode->timeout = fnode->mstimeout;
				  clnode->resolve = fnode->resolve;
				  clnode->transproxy = fnode->transproxy;
				  clnode->uid = fnode->uid;
				  clnode->gid = fnode->gid;
				  ID++;
				  clnode->ID = ID;
				  pthread_create (&thread, &threadattr,
						  (VOID *) phandleclient,
						  (VOID *) clnode);

				}
			      else
				{
				  // We just close immediately this connection attempt
				  if (msgsock)
				    close (msgsock);
				}

			      /* Unlock the list since we are done with it, and we want
			       * the threads to access it
			       */
			      pthread_mutex_unlock (&clist_lock);
			    }
			  else
			    {
			      //                            puts (" * Cannot accept connection");

			      // We still continue here
			    }
			}
		    }
		  else
		    {
		      syslog (LOG_NOTICE,
			      "Cannot listen to socket for port %d",
			      fnode->port);
		      ret = -1;
		    }
		}
	      else
		{
		  syslog (LOG_NOTICE, "Cannot bind socket for port %d",
			  fnode->port);
		  ret = -1;
		}

	      close (sock);
	    }
	  else
	    {
	      syslog (LOG_NOTICE, "Cannot create socket for port %d",
		      fnode->port);
	      ret = -1;
	    }

	  clist = freelist (clist);
	}
      else
	{
	  syslog (LOG_NOTICE, "Out of memory for port %d", fnode->port);
	  ret = -1;
	}

      exit (ret);
    }
  else if (pid > 0)
    {
      // Parent
      fnode->pid = pid;
      return (pid);
    }
  else if (pid == -1)
    {
      // -1, error
      return (0);
    }
}




// Daemon detach function
// ======================

BOOL make_daemon (VOID)
{
  int pid, fd;

  if ((getpid ()) != 1)
    {
      if ((pid = fork ()) != -1)
	{
	  if (pid > 0)
	    {

	      // Parent
	      exit (0);

	    }
	  else
	    {

	      // Child
	      setsid ();
	      chdir ("/");

	      if ((fd = open ("/dev/null", O_RDWR)) != -1)
		{
		  dup2 (fd, 0);
		  dup2 (fd, 1);
		  dup2 (fd, 2);
		  if (fd > 2)
		    close (fd);
		}

	      signal (SIGTTOU, SIG_IGN);
	      signal (SIGTTIN, SIG_IGN);
	      signal (SIGTSTP, SIG_IGN);
	      umask (0);

	      return (TRUE);

	    }
	}
    }

  return (FALSE);
}



BOOL write_pid (BYTE * pidpath)
{
  int fd;
  BYTE buf[64];

  if ((fd = open (pidpath, O_WRONLY | O_CREAT | O_TRUNC, 0644)) != -1)
    {
      sprintf (buf, "%ld\n", (LONG) getpid ());
      write (fd, buf, strlen (buf));
      close (fd);
    }

}


/* This function reads a valid (non-blank, non-comment) line from fh,
 * and parses the columns of that line, filling array with the pointers
 * to each trim string. It returns 0 if the number of columns of that line
 * is out of mincols,maxcols bounds, the number of columns of the line on
 * success, or -1 on EOF.
 */
LONG
readconfline (BYTE * str, FILE * fh, LONG mincols, LONG maxcols,
	      BYTE ** array)
{
  BYTE *ptr, *ptr2;
  LONG len, col;
  BOOL quoted;

  maxcols++;

  while (fgets (str, 1023, fh))
    {
      len = strlen (str);

      if (str[len - 1] == '\n')
	{
	  len--;
	  str[len] = 0;
	}
      if (str[len - 1] == '\r')
	{
	  len--;
	  str[len] = 0;
	}

      if (len)
	{
	  // Comment?
	  if (*str == '#' || *str == ';')
	    continue;

	  // No, so now format it's columns
	  col = 0;
	  ptr = str;

	  while ((*ptr) && (col < maxcols))
	    {

	      // Find first printable char
	      while (*ptr && (*ptr == ' ' || *ptr == '\t'))
		ptr++;

	      quoted = FALSE;
	      if (*ptr)
		{
		  ptr2 = ptr;

		  // Find next unprintable char
		  while (*ptr && *ptr != 34 && *ptr != ' ' && *ptr != '\t')
		    ptr++;

		  if (ptr != ptr2)
		    {
		      *ptr = 0;
		      ptr++;
		      array[col] = ptr2;
		      col++;
		    }
		  else
		    break;

		}
	      else
		break;

	    }

	  if ((col < maxcols) && (col >= mincols))
	    return (col);

	  return (0);

	}
    }

  return (-1);
}


/* This function starts a background thread, the root server. This is
 * done before using setuid() to become user nobody. Returns 0/NULL on error,
 * or returns the fd to send rootserver commands. It seems that pthreads
 * could not be used for this task, I thus had to fork() once again.
 */
rootdata *
start_rootserver (VOID)
{
  BOOL ok = FALSE;
  rootdata *rdata = NULL;
  int pid, fd;
  LONG len;
  BYTE buffer[1024], ipaddr[20], *ptr;

  // Allocate structure
  if ((rdata = (rootdata *) calloc (1, sizeof (rootdata))))
    {

      // First create our pipe (filedes[0]=read, filedes[1]=write)
      if (!(pipe (rdata->filedes)))
	{
	  // Pipe successfully created, start our rootserver thread
	  pid = fork ();
	  if (pid == 0)
	    {

	      // Child process, actual root thread. Detach...
	      setsid ();
	      chdir ("/");
	      if ((fd = open ("/dev/null", O_RDWR)) != -1)
		{
		  dup2 (fd, 0);
		  dup2 (fd, 1);
		  dup2 (fd, 2);
		  if (fd > 2)
		    close (fd);
		}
	      signal (SIGTTOU, SIG_IGN);
	      signal (SIGTTIN, SIG_IGN);
	      signal (SIGTSTP, SIG_IGN);
	      umask (0);

	      // Start actual activity
	      openlog ("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);

	      while ((len = fdgets (buffer, 1000, rdata->filedes[P_READ], -1))
		     != -2)
		{
		  syslog (LOG_NOTICE, "RootServer Request: \"%s\"", buffer);

		  /* The string we received should be in the format:
		   * n.n.n.n:p, make sure it is, isolate ip and port
		   */
		  ptr = buffer;
		  while (*ptr && *ptr != ':')
		    ptr++;

		  if (*ptr == ':')
		    {
		      *ptr = 0;
		      ptr++;

		      /* We isolated the ip and port parts,
		       * DENY request, modify firewall if not localhost
		       * address
		       */
		      strncpy (ipaddr, buffer, 19);
		      if ((strncmp (ipaddr, "127.", 4)))
			{
			  syslog (LOG_NOTICE,
				  "RootServer: DoS from [%s:%s] (DENYING)",
				  ipaddr, ptr);
			  sprintf (buffer,
				   "/sbin/ipchains -I input -l -s %s -j DENY",
				   ipaddr);
			  system (buffer);
			}
		    }
		}

	      closelog ();
	      exit (0);

	    }
	  if (pid > 0)
	    {

	      // Child started successfully, this is parent
	      rdata->pid = pid;
	      return (rdata);

	    }

	  // Error
	}

      free (rdata);

    }

  return (NULL);
}




// CLIENT SERVER (THREAD)
// ======================

// This is started by pthreads, it calls handleclients() and returns
VOID phandleclient (VOID * args)
{
  int ret, socket;
  clistnode *clnode = (clistnode *) args;
  list *clist = clnode->clist;
  pthread_mutex_t *clist_lock = clnode->clist_lock;

  /* Lock the list as we want to get the latest socket handle that points
   * to this client, which the main loop sets for us.
   */
  pthread_mutex_lock (clist_lock);
  socket = clnode->socket;
  pthread_mutex_unlock (clist_lock);

  // As a security measure, make sure we are wanted user and group
  setgid (clnode->gid);
  setuid (clnode->uid);

  /* Handle the client. We do not need to lock the clist during this, since
   * we are the only one which will actually delete it later on
   */
  ret = handleclient (socket, clnode);

  /* Decrease out connections number, if zero, delete our node. Here we lock
   * clist, since we might delete the node
   */
  pthread_mutex_lock (clist_lock);

  if (clnode->connections)
    clnode->connections--;

  if (!(clnode->connections))
    {
      makecurrent (clist, (lnode *) clnode);
      deletenode (clist);
    }

  pthread_mutex_unlock (clist_lock);
  pthread_exit (&ret);
}


// This function handles a client
int
handleclient (int fd, clistnode * clnode)
{
  struct hostent *hoste;
  struct in_addr ia;
  struct sockaddr_in addr_out;
  BYTE buffer[1024], ipaddr[20], toipaddr[20], hostname[255], uid[16];
  int ret = 0, tofd, selr, timeout;
  BYTE *tmp;
  LONG len, cnt;
  BOOL ok;
  BYTE unknown[] = "unknown";

  ULONG bytes_in, bytes_out, time_total;
  time_t time_start, time_end;
  LONG status;


  // Obtain IP address of client
  if ((tmp = inet_ntoa (clnode->client.sin_addr)))
    strncpy (ipaddr, tmp, 19);
  else
    strncpy (ipaddr, "0.0.0.0", 8);

  // Obtain IP address of server
  if ((tmp = inet_ntoa (clnode->toserver.sin_addr)))
    strncpy (toipaddr, tmp, 19);
  else
    strncpy (toipaddr, "0.0.0.0", 8);


  openlog ("mmtcpfwd", LOG_PID, LOG_AUTHPRIV);

  if (clnode->resolve)
    {
      // Obtain hostname of client
      ia.s_addr = inet_addr (ipaddr);
      if (
	  (hoste =
	   gethostbyaddr ((char *) &ia, sizeof (struct in_addr), AF_INET)))
	{
	  strncpy (hostname, hoste->h_name, 253);
	}
      else
	strncpy (hostname, unknown, 8);

      // Log user's ip and hostname
      syslog (LOG_NOTICE, "Connect: ID=0x%08X [%s:%d] - (%s) -> [%s:%d]",
	      clnode->ID, ipaddr, clnode->fromport, hostname, toipaddr,
	      clnode->toport);
    }
  else
    {
      syslog (LOG_NOTICE, "Connect: ID=0x%08X [%s:%d] -> [%s:%d]", clnode->ID,
	      ipaddr, clnode->fromport, toipaddr, clnode->toport);
    }

  bytes_in = bytes_out = 0;
  time_start = time (NULL);
  status = 0;

  ok = FALSE;
  if ((tofd = socket (AF_INET, SOCK_STREAM, 0)) > -1)
    {
      ok = TRUE;
    }
  else
    status = 8;

  if (ok)
    {
      ok = FALSE;

      if (clnode->transproxy)
	{
	  addr_out = clnode->client;
	  addr_out.sin_port = 0;

	  if (
	      (bind
	       (tofd, (struct sockaddr *) &addr_out,
		sizeof (struct sockaddr_in))) > -1)
	    {
	      ok = TRUE;
	    }
	  else
	    {
	      status = 9;
	    }
	}
      else
	{
	  ok = TRUE;
	}

    }

  if (ok)
    {
    }

  // Connect to IP and port we forward connection to (service)
  if (ok)
    {
      if (
	  (connect
	   (tofd, (struct sockaddr *) (void *) &clnode->toserver,
	    sizeof (clnode->toserver))) > -1)
	{

	  // Connected, now forward i/o in both directions

	  struct pollfd polldata[] = {
	    {fd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0},
	    {tofd, POLLIN | POLLPRI | POLLERR | POLLHUP | POLLNVAL, 0}
	  };
	  timeout = clnode->timeout;

	  ok = TRUE;
	  while (ok)
	    {

	      // Wait for fds to change state
	      polldata[0].revents = 0;
	      polldata[1].revents = 0;
	      selr = poll (polldata, 2, timeout);

	      if (selr > 0)
		{
		  /* State changed for at least one descriptor,
		   * find out which and perform appropriate actions
		   */

		  // First detect any error/connection losses
		  for (cnt = 0; cnt < 2; cnt++)
		    {
		      selr = polldata[cnt].revents;
		      if (selr & POLLERR || selr & POLLHUP || selr & POLLNVAL)
			{
			  ok = FALSE;
			  if (cnt == 0)
			    status = 1;
			  else
			    status = 2;
			  break;
			}
		    }

		  if (ok)
		    {

		      selr = polldata[0].revents;
		      if (selr & POLLIN || selr & POLLPRI)
			{
			  /* Bytes ready to read from client,
			   * forward them to actual service
			   */
			  if ((len = read (fd, buffer, sizeof (buffer) - 1)) <
			      1)
			    {
			      status = 3;
			      break;
			    }
			  if (len)
			    {
			      bytes_in += len;
			      if ((write (tofd, buffer, len)) != len)
				{
				  status = 5;
				  break;
				}
			    }
			}

		      selr = polldata[1].revents;
		      if (selr & POLLIN || selr & POLLPRI)
			{
			  /* Bytes available to read from service,
			   * forward them to client
			   */
			  if ((len = read (tofd, buffer, sizeof (buffer) - 1))
			      < 1)
			    {
			      status = 4;
			      break;
			    }
			  if (len)
			    {
			      bytes_out += len;
			      if ((write (fd, buffer, len)) != len)
				{
				  status = 6;
				  break;
				}
			    }
			}

		    }
		  else
		    break;	// Not ok ((-:

		}
	      else
		break;		// Timeout

	    }

	}
      else
	status = 7;

    }

  // Close session with service
  if (tofd != -1)
    close (tofd);

  // Close session with client
  close (fd);


  time_end = time (NULL);
  time_total = time_end - time_start;
  syslog (LOG_NOTICE,
	  "Closed: ID=0x%08X [%s:%d] - (In: %u, Out: %u, Seconds: %u, Status: %s)",
	  clnode->ID, ipaddr, clnode->fromport, bytes_in, bytes_out,
	  time_total, clnode->closed_status[status]);


  closelog ();
  return (ret);
}
